import pandas as pd
import numpy as np
from py_outliers_utils.outliers import outlier_identifier

def trim_outliers(dataframe, columns=None, identifier='IQR', method='trim'):
    """
    A function to generate outlier free dataset by imputing them with mean, median or trim entire row with outlier from dataset.
     ----------
    dataframe : pandas.core.frame.DataFrame
        The target dataframe where the function is performed.
    columns : list, default=None
        The target columns where the function needed to be performed. Default is None, the function will check all columns
    identifier : string
        The method of identifying outliers.
        - if "Z_score" : Use z-test with threshold of 3
        - if "IQR" : Use IQR (Inter Quantile range) to identify outliers
    method : string
        The method of dealing with outliers.
            - if "trim" :  remove completely rows with data points having outliers.
            - if "median" : replace outliers with median values
            - if "mean" : replace outliers with mean values
            
    Return
    -------
    pandas.core.frame.DataFrame
        a dataframe which the outlier has already process by the chosen method.
        
    Examples
    --------
    >>> import pandas as pd
    
    >>> df = pd.DataFrame({
    >>>    'SepalLengthCm' : [5.1, 4.9, 4.7, 5.5, 5.1, 50, 5.4, 5.0, 5.2, 5.3, 5.1],
    >>>    'SepalWidthCm' : [1.4, 1.4, 20, 2.0, 0.7, 1.6, 1.2, 1.4, 1.8, 1.5, 2.1],
    >>>    'PetalWidthCm' : [0.2, 0.2, 0.2, 0.3, 0.4, 0.5, 0.5, 0.6, 0.4, 0.2, 5]
    >>> })
    >>> trim_outliers(df, columns=['SepalLengthCm', 'SepalWidthCm', 'PetalWidthCm'],identifier='Z_score', method='trim')
    	 SepalLengthCm  	SepalWidthCm	   PetalWidthCm
    0	5.1	                1.4	                0.2
    1	4.9	                1.4	                0.2
    2	5.5	                2.0	                0.3
    3	5.1	                0.7	                0.4
    4	5.4	                1.2             	0.5
    5	5.0	                1.4	                0.6
    6	5.2	                1.8	                0.4
    7	5.3	                1.5	                0.2
    """

    # Handle dataframe type error (Check if dataframe is of type Pandas DataFrame)
    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError(f"passed dataframe is of type {type(dataframe).__name__}, should be DataFrame")
        
    # Handle empty dataframe or dataframe with all NAN
    if dataframe.empty or dataframe.dropna().empty:
        raise ValueError("passed dataframe is None")
    
    # Handle columns type error (Check if columns are None or type list)
    if not columns is None and not isinstance(columns, list):
        raise TypeError(f"passed columns is of type {type(columns).__name__}, should be list or NoneType")
    
    # Handle identifier type error (Check if identifier is of type str)
    if not isinstance(identifier, str):
        raise TypeError(f"passed identifier is of type {type(identifier).__name__}, should be string with value 'Z_score' or 'IQR'")
    
    # Handle identifier Value error (Check if identifier has value 'Z_score' or 'IQR')
    if identifier!='Z_score' and identifier!='IQR':
        raise ValueError("passed identifier should have value 'Z_score' or 'IQR'")
        
    # Handle method type error (Check if identifier method is 'trim', 'median' or 'mean')
    if method not in ("trim", "median", "mean"):
        raise Exception("The method must be -trim- or -median- or -mean-")


    dataframe_copy = dataframe.copy()
    
    if columns is None:
        columns = list(dataframe.columns)
    numeric_columns = dataframe[columns].select_dtypes('number').columns
    df = outlier_identifier(dataframe, columns=columns, identifier=identifier, return_df=True)
    if method == 'trim':
        outlier_index = df[df['outlier']==True].index
        dataframe_copy = dataframe_copy.drop(outlier_index)
    elif method == 'mean':
        col_list=[0]
        for col in numeric_columns:
            col_list[0] = col
            df_col = outlier_identifier(dataframe, columns=col_list, identifier=identifier, return_df = True)
            outlier_index = df_col[df_col['outlier']==True].index
            dataframe_copy.loc[outlier_index, col] = round(np.mean(dataframe[col]), 2)
    elif method == 'median':
        col_list=[0]
        for col in numeric_columns:
            col_list[0] = col
            df_col = outlier_identifier(dataframe, columns=col_list, identifier=identifier, return_df = True)
            outlier_index = df_col[df_col['outlier']==True].index
            dataframe_copy.loc[outlier_index, col] = round(np.median(dataframe[col]), 2)
    return dataframe_copy


import pandas as pd
import altair as alt

def visualize_outliers(dataframe, columns=None, type='violin'):
    """
    A function that plot the distribution of the given data.
    
    Parameters
    ----------
    dataframe : pandas.core.frame.DataFrame
        The target dataframe where the function is performed.
    
    columns : list, default=None
        The target columns where the function needed to be performed. Default is None, the function will check all columns.
    
    type : string, default='violin'
        The method of plotting the distribution.
        - if "violin" : Return a violin plot with boxplot layer
        - if "boxplot" : Return a boxplot


    Returns
    -------
    altair.vegalite.v4.api.Chart
        an altair plot with data distribution.
    
    Examples
    --------
    >>> import pandas as pd
    >>> import altair as alt
        
    >>> df = pd.DataFrame({
    >>>    'SepalLengthCm' : [0.1, 4.9, 52.7, 5.5, 5.1, 50, 5.4, 179.0, 5.2, 5.3, 5.1],
    >>>    'SepalWidthCm' : [1.4, 1.4, 20, 2.0, 0.7, 1.6, 1.2, 14, 1.8, 1.5, 2.1],
    >>>    'PetalWidthCm' : [0.2, 0.2, 0.2, 0.3, 0.4, 0.5, 0.5, 0.6, 0.4, 0.2, 5]
    >>> })

    >>> visualize_outliers(df, columns=['SepalLengthCm', 'SepalWidthCm'])
    """

    ## Handle dataframe type error (Check if dataframe is of type Pandas DataFrame)
    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError(f"passed dataframe is of type {type(dataframe).__name__}, should be DataFrame")
        
    ## Handle empty dataframe or dataframe with all NAN
    if dataframe.empty or dataframe.dropna().empty:
        raise ValueError("passed dataframe is None")
    
    ## Handle columns type error (Check if columns are None or type list)
    if not columns is None and not isinstance(columns, list):
        raise TypeError(f"passed columns is of type {type(columns).__name__}, should be list or NoneType")
    
    ## Handle type Value error (Check if type has value 'violin' or 'boxplot')
    if type!='violin' and type!='boxplot':
        raise ValueError("passed type should have value 'violin' or 'boxplot'")
    
    ## Select given columns
    if columns is None:
        df = dataframe
    else:
        df = dataframe[columns]
    
    ## Select numeric columns
    dfnumeric = df._get_numeric_data() 
    
    ## Melt dataframe for plot
    dfmelt = dfnumeric.melt()  
    
    ## Plot according type
    if type == 'violin':
        
        ## The boxplot layer
        boxplot = alt.Chart().mark_boxplot(color='black').encode(
                            alt.Y('value')
                    ).properties(width=100)
        
        ## The violinplot layer
        violin = alt.Chart().transform_density(
                'value',
                as_=['value', 'density'],
                groupby=['variable']
            ).mark_area(orient='horizontal').encode(
                y=alt.Y('value'),
                color=alt.Color('variable:N', legend=None),
                x=alt.X(
                    'density:Q',
                    stack='center',
                    impute=None,
                    title=None,
                    scale=alt.Scale(nice=False,zero=False),
                    axis=alt.Axis(labels=False, values=[0], grid=False, ticks=True),
                ),
            )
        
        ## Overlay two layers
        violinbox = alt.layer(violin, boxplot, data=dfmelt).facet('variable:N', columns=5).resolve_scale(x=alt.ResolveMode("independent"))
        return violinbox
    
    else:
        boxplot = alt.Chart(dfmelt).mark_boxplot().encode(
                    alt.X(title = None),
                    alt.Y("value"),
                    alt.Color("variable", legend = None) 
                ).properties(
                    width=100
                ).facet('variable:N', columns=4)
        return boxplot



import pandas as pd
import numpy as np
def outlier_identifier(dataframe, columns=None, identifier = 'IQR', return_df=False):
    """
    A function that identify and summarize the count and range of based on the method the user choose
    Parameters
    ----------
    dataframe : pandas.core.frame.DataFrame
        The target dataframe where the function is performed.
    columns : list, default=None
        The target columns where the function needed to be performed. Default is None, the function will check all columns
    identifier : string, default='IQR'
        The method of identifying outliers.
        - if "Z_score" : Use z-test with threshold of 3
        - if "IQR" : Use IQR (Inter Quantile range) to identify outliers (default)
    return_df : bool, default=False
        Can be set to True if want output as dataframe identified with outliers in rows

    Returns
    -------
    pandas.core.frame.DataFrame 
    (a dataframe with the summary of the outlier identified by the method) if return_df = False , 
    (a dataframe with additional column having if row has outlier or not) if return_df = True
    
    Examples
    --------
    >>> import pandas as pd
        
    >>> df = pd.DataFrame({
    >>>    'SepalLengthCm' : [5.1, 4.9, 4.7, 5.5, 5.1, 50, 54, 5.0, 5.2, 5.3, 5.1],
    >>>    'SepalWidthCm' : [1.4, 1.4, 20, 2.0, 0.7, 1.6, 1.2, 1.4, 1.8, 1.5, 2.1],
    >>>    'PetalWidthCm' : [0.2, 0.2, 0.2, 0.3, 0.4, 0.5, 0.5, 0.6, 0.4, 0.2, 5]
    >>> })
    >>> outlier_identifier(df)
    	                SepalLengthCm SepalWidthCm PetalWidthCm
    outlier_count                  2            1            1
    outlier_percentage        18.18%        9.09%        9.09%
    mean                       13.63         3.19         0.77
    median                       5.1          1.5          0.4
    std                        18.99         5.59         1.41
    lower_range                  NaN          NaN          NaN
    upper_range         (50.0, 54.0)         20.0          5.0
    """

    # Handle dataframe type error (Check if dataframe is of type Pandas DataFrame)
    if not isinstance(dataframe, pd.DataFrame):
        raise TypeError(f"passed dataframe is of type {type(dataframe).__name__}, should be DataFrame")
        
    # Handle empty dataframe or dataframe with all NAN
    if dataframe.empty or dataframe.dropna().empty:
        raise ValueError("passed dataframe is None")
    
    # Handle columns type error (Check if columns are None or type list)
    if not columns is None and not isinstance(columns, list):
        raise TypeError(f"passed columns is of type {type(columns).__name__}, should be list or NoneType")
    
    # Handle identifier type error (Check if identifier is of type str)
    if not isinstance(identifier, str):
        raise TypeError(f"passed identifier is of type {type(identifier).__name__}, should be string with value 'Z_score' or 'IQR'")
    
    # Handle identifier Value error (Check if identifier has value 'Z_score' or 'IQR')
    if identifier!='Z_score' and identifier!='IQR':
        raise ValueError("passed identifier should have value 'Z_score' or 'IQR'")
        
    # Handle return_df type error (Check if identifier is of type bool)
    if not isinstance(return_df, bool):
        raise TypeError(f"passed return_df is of type {type(return_df).__name__}, should be bool with value as True or False")

    if columns is None:
        columns = dataframe.columns
    df_filtered = dataframe[columns]
    numeric_columns = df_filtered.select_dtypes('number').columns
    df_selected = df_filtered[numeric_columns]
    df_selected['outlier'] = False
    
    output = pd.DataFrame(columns=numeric_columns, index=['outlier_count', 'outlier_percentage', 'mean', 'median', 'std', 'lower_range', 'upper_range'])
    if identifier == 'Z_score':
        if return_df == False:
            for col in numeric_columns:
                std = df_selected[col].std()
                mean = df_selected[col].mean()
                col_outliers = df_selected[np.abs(df_selected[col] - mean) > std*3]
                output.loc['outlier_count', col] = len(col_outliers)
                output.loc['outlier_percentage', col] = str(round(len(col_outliers) * 100 / len(df_selected), 2)) + '%'
                output.loc['mean', col] = round(mean, 2)
                output.loc['median', col] = round(np.percentile(df_selected[col], 50), 2)
                output.loc['std', col] = round(std, 2)

                low_range = col_outliers[col_outliers[col] < -std*3]
                if not low_range.empty:
                    if len(low_range) == 1:
                        output.loc['lower_range', col] = min(low_range[col])
                    else:
                        output.loc['lower_range', col] = (min(low_range[col]), max(low_range[col]))
                upper_range = col_outliers[col_outliers[col] > std*3]
                if not upper_range.empty:
                    if len(upper_range) == 1:
                        output.loc['upper_range', col] = min(upper_range[col])
                    else:
                        output.loc['upper_range', col] = (min(upper_range[col]), max(upper_range[col]))
            return output
        else:
            for col in numeric_columns:
                std = df_selected[col].std()
                mean = df_selected[col].mean()
                df_selected.loc[np.abs((df_selected[col] - mean)) > std*3, 'outlier'] = True
            return df_selected


    else:
        if return_df == False:
            for col in numeric_columns:
                std = df_selected[col].std()
                mean = df_selected[col].mean()
                iqr = np.percentile(df_selected[col], 75) - np.percentile(df_selected[col], 25)
                col_outliers = df_selected[np.abs((df_selected[col] - mean) / std) > 1.5*iqr]
                output.loc['outlier_count', col] = len(col_outliers)
                output.loc['outlier_percentage', col] = str(round(len(col_outliers) * 100 / len(df_selected), 2)) + '%'
                output.loc['mean', col] = round(mean, 2)
                output.loc['median', col] = round(np.percentile(df_selected[col], 50), 2)
                output.loc['std', col] = round(std, 2)
                low_range = col_outliers[col_outliers[col] < -1.5*iqr]
                if not low_range.empty:
                    if len(low_range) == 1:
                        output.loc['lower_range', col] = min(low_range[col])
                    else:
                        output.loc['lower_range', col] = (min(low_range[col]), max(low_range[col]))
                upper_range = col_outliers[col_outliers[col] > 1.5*iqr]
                if not upper_range.empty:
                    if len(upper_range) == 1:
                        output.loc['upper_range', col] = min(upper_range[col])
                    else:
                        output.loc['upper_range', col] = (min(upper_range[col]), max(upper_range[col]))
            return output

        else:
            for col in numeric_columns:
                std = df_selected[col].std()
                mean = df_selected[col].mean()
                iqr = np.percentile(df_selected[col], 75) - np.percentile(df_selected[col], 25)
                df_selected.loc[np.abs((df_selected[col] - mean) / std) > 1.5*iqr, 'outlier'] = True
            
            return df_selected


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['py_outliers_utils']

package_data = \
{'': ['*']}

install_requires = \
['altair>=4.2.0,<5.0.0', 'pandas>=1.4.0,<2.0.0']

setup_kwargs = {
    'name': 'py-outliers-utils',
    'version': '1.2.7',
    'description': 'Dealing with outliers',
    'long_description': '# py_outliers_utils \n\n[![ci-cd](https://github.com/UBC-MDS/py_outliers_utils/actions/workflows/ci-cd.yml/badge.svg)](https://github.com/UBC-MDS/py_outliers_utils/actions/workflows/ci-cd.yml)\n\n## Overview\nAs data rarely comes ready to be used and analyzed for machine learning right away, this package aims to help speed up the process of cleaning and doing initial exploratory data analysis specific to outliers. The package focuses on the tasks of identifying univariate outliers, providing summary of outliers like count, range of outliers, visualize them and giving functionality to remove them from data.\n\n## Installation\n\n```bash\n$ pip install py_outliers_utils\n```\n## Functions\nThe three functions contained in this package are as follows:\n\n- `outlier_identifier`: A function to identify outliers in the dataset and provide their summary as an output\n- `visualize_outliers`: A function to generate the eda plots highlighting outliers providing additional functionality to visualize them\n- `trim_outliers`: A function to generate outlier free dataset by imputing them with mean, median or trim entire row with outlier from dataset.\n\n## Our Place in the Python Ecosystem\nWhile Python packages with similar functionalities exist, this package aims to provide summary, visualization of outliers in a single package with an additional functionality to generate outlier-free dataset. Few packages with similar functionality are as follows:\n\n- [pyod](https://pypi.org/project/pyod/)\n- [python-outlier](https://pypi.org/project/python-outlier/)\n\n## Usage\n\nTo import the package and check the version:\n\n```bash\nimport py_outliers_utils\n\nprint(py_outliers_utils.__version__)\n```\n\n`py_outliers_utils` can be used to deal with the outliers in a dataset and plot the distribution of the dataset. More information can be found in the [documentation](https://py-outliers-utils.readthedocs.io/en/latest/example.html).\n\n## Contributing\n\nThis package is authored by Karanpreet Kaur, Linhan Cai, Qingqing Song as part of the course project in DSCI-524 (UBC-MDS program). You can see the list of all contributors in the contributors tab.\n\nWe welcome and recognize all contributions. If you wish to participate, please review our [Contributing guidelines](CONTRIBUTING.md)\n\n## License\n\n`py_outliers_utils` is licensed under the terms of the MIT license.\n\n## Credits\n\n`py_outliers_utils` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n',
    'author': 'scarlqq',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
